import requests
import json
from typing import List, Dict, Optional, Union
from urllib.parse import urlencode
from http import HTTPStatus


class Api:
    DOMAINS = {'es': 'www.emagister.com',
               'uk': 'www.emagister.co.uk',
               'fr': 'www.emagister.fr'}

    BASE_API_URL = 'https://{}/api/'
    MAX_NUM_OF_CONSECUTIVE_ERRORS = 5
    ENCODING = 'utf-8'

    def __init__(self,
                 country: str = 'es',
                 version: str = '1.0',
                 page_size: int = 25,
                 silent: bool = True):
        """Api class constructor
        :param country: Country site
        :param version: Api version
        :param page_size: Number of records per page
        :param silent: Whether or not messages will be printed on the screen
        """
        self.country = country
        self.version = version
        self.page_size = page_size
        self.records = []
        self.silent = silent
        self.errors = 0

    def get(self,
            subset: Optional[Union[List[str], Dict]] = None,
            filters: Optional[Dict] = None,
            max_records: Optional[int] = None,
            url: Optional[str] = None) -> List:

        """ Get courses collection from API
        :param subset: Subset of fields to retrieve. If None, all fields will be retrieved.
        :param filters: Dictionary of filters to apply. See https://www.emagister.com/api/doc#get--api-{version}-courses
        :param max_records: Maximum number of records to retrieve. If None, all records will be retrieved.
        :param url: Api endpoint
        :return: A list of dictionaries representing records
        """
        raise NotImplementedError

    @staticmethod
    def __record_data__(cls, record: Dict, subset: Optional[Union[List[str], Dict]] = None) -> Dict:
        if not subset:
            return record

        data = {}

        if isinstance(subset, list):
            for key in subset:
                data[key] = record[key]

            return data

        if isinstance(subset, dict):
            for key, value in subset.items():
                data[key] = record[value]

            return data

        raise Exception('subset must be of type Dictionary or List')

    def __build_url__(self, resource: str, filters=None, page: int = 1) -> str:
        base_api_url = self.BASE_API_URL.format(self.DOMAINS[self.country])

        parameters = 'page={}&size={}'.format(page, self.page_size)

        if filters:
            parameters = '{}&{}'.format(parameters, urlencode(filters))

        return '{}{}/{}?{}'.format(base_api_url, self.version, resource, parameters)

    def __read__(self, url: str) -> Dict:
        response = requests.get(url)

        if response.status_code == HTTPStatus.OK:
            if not self.silent:
                print('GET: {} [{}]'.format(url, response.status_code))

            self.errors = 0

            return json.loads(response.content.decode(self.ENCODING))

        if not self.silent:
            print('GET: {} [{}]'.format(url, response.status_code))

        self.errors += 1
        if self.errors == self.MAX_NUM_OF_CONSECUTIVE_ERRORS:
            raise Exception('Maximum number of consecutive errors reached')

    def __add_record__(self, record: Dict):
        self.records.append(record)

    def __get_records__(self) -> List:
        records = self.records
        self.records = []

        return records


class Courses(Api):
    def get(self,
            subset: Optional[Union[List[str], Dict]] = None,
            filters: Optional[Dict] = None,
            max_records: Optional[int] = None,
            url: Optional[str] = None) -> List:

        if not url:
            url = self.__build_url__('courses', filters=filters)

        courses = self.__read__(url)
        total_records = max_records if max_records else courses['total_count']

        next_url = courses['_links']['next']['href']

        for course in courses['courses']:
            course_data = self.__record_data__(course, subset)

            self.__add_record__(course_data)

            if max_records and len(self.records) >= max_records:
                return self.__get_records__()

        if not self.silent:
            print('Records: {}/{}'.format(len(self.records), total_records))

        if next_url:
            return self.get(subset, max_records=max_records, url=next_url)

        return self.__get_records__()
